/* $Id: axprt_unix.C,v 1.27 1999/12/01 02:09:12 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "arpc.h"
#include "vector.h"

#ifdef MAINTAINER
bool axprt_unix_spawn_connected;
#endif /* MAINTAINER */
pid_t axprt_unix_spawn_pid;

void
axprt_unix::clone (ref<axprt_clone> x)
{
  assert (pktsize >= x->pktsize); // sic - allow no possibility of failure
  assert (!x->ateof ());

  sendfd (x->takefd ());
  assert (x->pktlen >= 4);
  send (x->pktbuf + 4, x->pktlen - 4, NULL);
}

axprt_unix::~axprt_unix ()
{
  while (!fdrecvq.empty ())
    close (fdrecvq.pop_front ());
}

void
axprt_unix::sendit (int f, bool closeit)
{
  if (fd >= 0) {
    /* XXX - This code is not correct.  If !allow_recvfd, we
     * presumably do not trust the remote end not to do funny stuff.
     * In particular, the remote end could set things up so that there
     * is no room for the one byte synchronously transmitted by
     * ::sendfd.  This would allow the remote end to hang this
     * process.  Instead, we just drop the file descriptor.
     *
     * This solotuion here could in theory cause incorrect behavior if
     * flow control gets unlucky.  In practice, however, RPC's by
     * untrusted processes requesting file descriptors have short
     * replies and occur without other outstanding RPC traffic.  Thus,
     * this won't cause a problem in current use of the code.
     *
     * There is also potential for deadlock between two trusted
     * processes (allow_recvfd on both) when each is sending the other
     * a file descriptor.  This cannot occur in practice (because of
     * the scheduling of the event loop) unless one event in each
     * process causes large RPC's into the other with a passed file
     * descriptor in between.
     */
    timeval ztv = { 0, 0 };
    if (!allow_recvfd && fdwait (fd, selwrite, &ztv) <= 0)
      warn ("sendfd: cannot send\n");
    else if (::sendfd (fd, f) < 0)
      warn ("sendfd: %m\n");
  }
  if (closeit)
    close (f);
}

void
axprt_unix::recvbreak ()
{
  if (!allow_recvfd) {
    axprt_stream::recvbreak ();
    return;
  }
  int nfd = ::recvfd (fd);
  if (nfd < 0)
    warn ("recvfd: %m\n");
  else if (fdrecvq.size () >= 2) {
    close (nfd);
    warn ("recvfd: too many file descriptors\n");
  }
  else
    fdrecvq.push_back (nfd);
}

int
axprt_unix::recvfd ()
{
  if (fdrecvq.empty ())
    return -1;
  return fdrecvq.pop_front ();
}

void
axprt_unix::sendfd (int sfd, bool closeit)
{
  sendbreak (wrap (this, &axprt_unix::sendit, sfd, closeit));
}

#if 1 /* Hmm... in practice this isn't always necessary */
/* Unfortunately, we need to "sync" up our reads to deal with file
 * descriptor passing.  That means we never read beyond the length of
 * the next packet (as that length may be a 0 i.e. "break" followed by
 * a file descriptor).  */
size_t
axprt_unix::readsize ()
{
  if (pktlen < 4)
    return 4 - pktlen;
  u_int32_t psize = getint (pktbuf) & 0x7fffffffU;
  return min<size_t> (bufsize - pktlen, psize + 8 - pktlen);
}
#endif

ref<axprt_unix>
axprt_unix::alloc (int f, size_t ps)
{
  ref<axprt_unix> x = New refcounted<axprt_unix> (f, ps);

  if (!isunixsocket (f)) {
    warn ("axprt_unix::alloc(%d): not unix domain socket\n", f);
    x->fail ();
  }
  return x;
}

#ifdef MAINTAINER
static ptr<axprt_unix>
tryconnect (str path, const char *arg0, u_int ps)
{
  const char *prog = strrchr (path, '/');
  if (!prog)
    panic ("tryconnect: path '%s' has no '/'\n", path.cstr ());
  prog++;

  if (builddir) {
    const char *a0;
    if (arg0) {
      if ((a0 = strrchr (arg0, '/')))
	a0++;
      else
	a0 = arg0;
    }
    else
      a0 = prog;

    str np = strbuf ("%s/.%s",
		     buildtmpdir ? buildtmpdir.cstr () : builddir.cstr (),
		     a0);
    return axprt_unix_connect (np, ps);
  }
  return NULL;
}
#endif /* MAINTAINER */

static ptr<axprt_unix>
axprt_unix_dospawnv (str path, const vector<str> &avs,
		     size_t ps, cbv::ptr postforkcb, bool async)
{
  axprt_unix_spawn_pid = -1;
  vector<char *> av;
  if (!ps)
    ps = axprt_stream::defps;

  // path = fix_exec_path (path);
#ifdef MAINTAINER
  if (ptr<axprt_unix> x = tryconnect (path, avs[0], ps)) {
    axprt_unix_spawn_connected = true;
    return x;
  }
  axprt_unix_spawn_connected = false;
#endif /* MAINTAINER */

  for (const str *s = avs.base (), *e = avs.lim (); s < e; s++)
    av.push_back (const_cast<char *> (s->cstr ()));
  av.push_back (NULL);

  int fds[2];
  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
    warn ("socketpair: %m\n");
    return NULL;
  }
  close_on_exec (fds[0]);
  pid_t pid = (async ? &aspawn : &spawn) (path, av.base (), fds[1], 1, 2,
					  postforkcb);
  axprt_unix_spawn_pid = pid;
  close (fds[1]);
  if (pid < 0) {
    close (fds[0]);
    return NULL;
  }

  // return axprt_unix::alloc (fds[0], ps);
  // XXX - extra complexity to work around g++ bug
  ref<axprt_unix> x = axprt_unix::alloc (fds[0], ps);
  return x;
}

ptr<axprt_unix>
axprt_unix_spawnv (str path, const vector<str> &avs,
		   size_t ps, cbv::ptr postforkcb)
{
  ptr<axprt_unix> x = axprt_unix_dospawnv (path, avs, ps, postforkcb, false);
  return x;
}
ptr<axprt_unix>
axprt_unix_aspawnv (str path, const vector<str> &avs,
		    size_t ps, cbv::ptr postforkcb)
{
  ptr<axprt_unix> x = axprt_unix_dospawnv (path, avs, ps, postforkcb, true);
  return x;
}

ptr<axprt_unix>
axprt_unix_spawn (str path, size_t ps, char *arg0, ...)
{
  axprt_unix_spawn_pid = -1;
  vector<char *> av;
  if (!ps)
    ps = axprt_stream::defps;

  if (arg0) {
    va_list ap;
    va_start (ap, arg0);
    av.push_back (arg0);
    while (av.push_back (va_arg (ap, char *)))
      ;
    va_end (ap);
  }
  else {
    av.push_back (const_cast<char *> (path.cstr ()));
    av.push_back (NULL);
  }

  // path = fix_exec_path (path);
#ifdef MAINTAINER
  if (ptr<axprt_unix> x = tryconnect (path, arg0, ps)) {
    axprt_unix_spawn_connected = true;
    return x;
  }
  axprt_unix_spawn_connected = false;
#endif /* MAINTAINER */

  int fds[2];
  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
    warn ("socketpair: %m\n");
    return NULL;
  }
  close_on_exec (fds[0]);
  pid_t pid = spawn (path, av.base (), fds[1]);
  axprt_unix_spawn_pid = pid;
  close (fds[1]);
  if (pid < 0) {
    close (fds[0]);
    return NULL;
  }

  // return axprt_unix::alloc (fds[0], ps);
  // XXX - extra complexity to work around g++ bug
  ref<axprt_unix> x = axprt_unix::alloc (fds[0], ps);
  return x;
}

ptr<axprt_unix>
axprt_unix_connect (const char *path, size_t ps)
{
  int fd = unixsocket_connect (path);
  if (fd < 0)
    return NULL;
  return axprt_unix::alloc (fd, ps);
}

#ifdef MAINTAINER
ptr<axprt_unix>
axprt_unix_accept (const char *path, size_t ps)
{
  mode_t m = umask (0);
  int fd = unixsocket (path);
  if (fd < 0) {
    warn ("unixsocket: %m\n");
    umask (m);
    return NULL;
  }
  umask (m);

  sockaddr_un sun;
  socklen_t len = sizeof (sun);
  bzero (&sun, sizeof (sun));
  int afd = -1;
  if (!listen (fd, 1))
    afd = accept (fd, (sockaddr *) &sun, &len);
  unlink (path);
  close (fd);

  if (afd < 0) {
    warn ("%s: %m\n", path);
    return NULL;
  }
  return axprt_unix::alloc (afd, ps);
}
#endif /* MAINTAINER */

ptr<axprt_unix>
axprt_unix_stdin (size_t ps)
{
  ptr<axprt_unix> x = axprt_unix::alloc (0);
#ifdef MAINTAINER
  if (x->ateof () && builddir) {
    str np = strbuf ("%s/.%s",
		     buildtmpdir ? buildtmpdir.cstr () : builddir.cstr (),
		     progname.cstr ());
    x = axprt_unix_accept (np, ps);
  }
#endif /* MAINTAINER */
  if (x && !x->ateof ())
    return x;
  warn ("axprt_unix_stdin: %m\n");
  return NULL;
}
