/* $Id: usrinfo.C,v 1.59 2000/01/05 20:59:07 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sfscd.h"
#include "sfsagent.h"

const time_t agent_timeout = 60;

qhash<sfs_aid, ref<usrinfo> > &usrtab (*New qhash<sfs_aid, ref<usrinfo> >);
static ptr<afslink> revokedlink;

sfs_aid
cdgetaid (const authunix_parms *aup)
{
  if (root_is_nfsmounter && !aup->aup_uid)
    return sfsaid_sfs;
  return aup2aid (aup);
}

static void
nullccb_fn (clnt_stat)
{
}
static aclnt_cb nullccb (wrap (nullccb_fn));
static void
flushuser (sfs_aid aid, cdaemon *cdp)
{
  if (aclnt *c = cdp->c)
    c->call (SFSCDPROC_FLUSHAUTH, &aid, NULL, nullccb);
}

struct newagent {
  const sfs_aid aid;
  ref<axprt_unix> ax;
  ref<asrv> as;

  newagent (ref<axprt_unix> x, const authunix_parms *aup)
    : aid (cdgetaid (aup)), ax (x),
      as (asrv::alloc (ax, agent_prog_1, wrap (this, &newagent::dispatch)))
    {}
  void dispatch (svccb *);
};

static void
agent_default (svccb *sbp, sfs_aid aid)
{
  switch (sbp->proc ()) {
  case AGENT_NULL:
    sbp->reply (NULL);
    break;
  case AGENT_REVOKE:
    revocation::alloc (*sbp->template getarg<sfs_pathrevoke> ());
    sbp->reply (NULL);
    break;
  case AGENT_RNDSEED:
    {
      sfsagent_seed seed;
      rnd.getbytes (seed.base (), seed.size ());
      sbp->reply (&seed);
      break;
    }
  case AGENT_AIDALLOC:
    {
      if (sfs_specaid (aid)) {
	sbp->replyref (EINVAL);
	break;
      }
      u_int32_t uid = aid & 0xffffffff;
      u_int32_t gid = sfs_resvgid_start + sfs_resvgid_count;
      while (gid-- > sfs_resvgid_start)
	if (!usrtab[sfs_mkaid (uid, gid)])
	  break;
      if (gid < sfs_resvgid_start)
	gid = sfs_badgid;
      sbp->reply (&gid);
      break;
    }
  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

void
newagent::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }

  switch (sbp->proc ()) {
  case AGENT_KILLSTART:
    {
      if (aid == sfsaid_sfs || aid == sfsaid_nobody)
	sbp->replyref (EINVAL);
      ref<usrinfo> u (New refcounted<usrinfo> (aid));
      u->sh = as;
      u->ch = aclnt::alloc (ax, agentcb_prog_1);
      u->x = ax;
      usrtab.insert (u->aid, u);
      as->setcb (wrap (u.get (), &usrinfo::dispatch));
      delete this;
      sbp->replyref (0);
      break;
    }

  case AGENT_START:
    if (aid == sfsaid_sfs || aid == sfsaid_nobody)
      sbp->replyref (EINVAL);
    else if (usrtab[aid])
      sbp->replyref (EBUSY);
    else {
      ref<usrinfo> u (New refcounted<usrinfo> (aid));
      u->sh = as;
      u->ch = aclnt::alloc (ax, agentcb_prog_1);
      u->x = ax;
      usrtab.insert (u->aid, u);
      as->setcb (wrap (u.get (), &usrinfo::dispatch));
      delete this;
      sbp->replyref (0);
    }
    break;

  case AGENT_KILL:
    if (usrtab[aid]) {
      usrtab.remove (aid);
      sbp->replyref (0);
    }
    else
      sbp->replyref (ESRCH);
    break;

  case AGENT_GETAGENT:
    if (ptr<usrinfo> u = usrtab[aid]) {
      if (u->clonelock) {
	sbp->replyref (EAGAIN);
	break;
      }
      int fds[2];
      if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
	sbp->replyref (errno);
	break;
      }
      u->clonelock = true;
      ax->sendfd (fds[0]);
      u->x->sendfd (fds[1]);
      u->ch->call (AGENTCB_CLONE, NULL, NULL, aclnt_cb_null);
      sbp->replyref (0);
      u->x->setwcb (wrap (ref<axprt_stream> (ax), &axprt_stream::setwcb,
			  wrap (u, &usrinfo::clonelock_clear)));
    }
    else
      sbp->replyref (ESRCH);
    break;

  default:
    agent_default (sbp, aid);
  }

}

ptr<usrinfo>
usrinfo::lookup (const authunix_parms *aup)
{
  return usrtab[cdgetaid (aup)];
}

static void
authreq_cb (svccb *sbp, ref<sfsagent_auth_res> resp, clnt_stat err)
{
  if (err)
    resp->set_authenticate (false);
  sbp->reply (resp);
}

void
usrinfo::authreq (sfscd_authreq *ar, svccb *sbp)
{
  ref<sfsagent_auth_res> resp = New refcounted<sfsagent_auth_res>;
  switch (ar->type) {
  case AGENTCB_AUTHINIT:
    ch->timedcall (agent_timeout, AGENTCB_AUTHINIT, ar->init.addr (), resp,
		   wrap (authreq_cb, sbp, resp));
    break;
  case AGENTCB_AUTHMORE:
    ch->timedcall (agent_timeout, AGENTCB_AUTHMORE, ar->more.addr (), resp,
		   wrap (authreq_cb, sbp, resp));
    break;
  default:
    warn ("usrinfo::authreq: bad type\n");
    sbp->replyref (sfsagent_auth_res (false));
    break;
  }
}

ref<afsdir>
userdir (const svccb *sbp)
{
  if (!sbp)
    return afs_naroot;
  const sfs_aid aid = cdgetaid (sbp->getaup ());
  if (aid == sfsaid_sfs)
    return afs_sfsroot;
  if (usrinfo *u = usrtab[aid])
    return u->root;
  return afs_naroot;
}

static void
login_user_cb (ref<afsdir> d, str name, ref<delaypt> dpt,
	       struct srvinfo *si, sfsagent_revoked_res *_rp, clnt_stat err)
{
  if (!revokedlink)
    revokedlink = afslink::alloc (":REVOKED:");

  // XXX - here we should log the user in
  auto_ptr<sfsagent_revoked_res> rp (_rp);

  d->unlink (name);
  if (err == RPC_PROCUNAVAIL)
    err = RPC_SUCCESS;

  if (err) {
    dpt->setres (NFSERR_IO);
    return;
  }

  switch (rp->type) {
  case REVOCATION_NONE:
    if (afsnode *e = afs_sfsroot->lookup (name, NULL))
      d->link (e, name);
    else {
      dpt->setres (NFSERR_NOENT);
      return;
    }
    break;
  case REVOCATION_BLOCK:
    d->link (revokedlink, name);
    break;
  case REVOCATION_CERT:
    if (ptr<afsnode> e = revocation::alloc (*rp->cert))
      d->link (e, name);
    else
      d->link (revokedlink, name);
    break;
  }

  dpt->setres (strbuf () << sfsroot << "/" << name);
}

static void
login_user (ptr<usrinfo> u, ref<afsdir> d, str name, ref<delaypt> dpt,
	    struct srvinfo *si)
{
  sfsagent_revoked_res *rp = New sfsagent_revoked_res;
  if (u)
    u->ch->timedcall (agent_timeout, AGENTCB_REVOKED, &name, rp,
		      wrap (login_user_cb, d, name, dpt, si, rp));
  else
    login_user_cb (d, name, dpt, si, rp, RPC_SUCCESS);
}

void
afsdir_unlink (ref<afsdir> d, str name)
{
  d->unlink (name);
}

static void
automount_cb (ptr<usrinfo> u, ptr<afsdir> d, str name, ref<delaypt> dpt,
	      struct srvinfo *si, int err)
{
  if (si && !err)
    login_user (u, d, name, dpt, si);
  else {
    dpt->setres (nfsstat (err));
    // dpt->setres (NFSERR_NOENT);
    timecb (timenow + 2, wrap (afsdir_unlink, d, name));
  }
}

void
agentlookup_cb (ref<afsusrdir> d, str name, ref<delaypt> dpt,
		sfsagent_lookup_res *resp, clnt_stat err)
{
  d->unlink (name);
  if (err || !resp->makelink || !resp->path->len ()) {
    d->mkulink (NULL, name);
    dpt->setres (NFSERR_NOENT);
  }
  else {
    d->mkulink (*resp->path, name);
    dpt->setres (*resp->path);
  }
  delete resp;
}

afsnode *
automount (svccb *sbp, str name)
{
  assert (!isnfsmounter (sbp->getaup ()));

  bool remote = sfs_parsepath (name);
  ptr<usrinfo> u = usrinfo::lookup (sbp->getaup ());
  if (!remote && (!nameok (name) || !u || u->root->negcache[name]))
    return NULL;

  srvinfo *si = remote ? srvinfo::lookup (name) : NULL;
  if (si && si->error)
    return NULL;

  ref<afsdir> d (u ? implicit_cast<ref<afsdir> > (u->root) : userdir (sbp));
  d->unlink (name);
  ref<delaypt> dpt = delaypt::alloc ();
  d->link (dpt, name);

  if (remote) {
    if (si)
      login_user (u, d, name, dpt, si);
    else
      srvinfo::alloc (name, wrap (automount_cb, u, d, name, dpt));
  }
  else {
    sfsagent_lookup_res *resp = New sfsagent_lookup_res;
    u->ch->timedcall (agent_timeout, AGENTCB_LOOKUP, &name, resp,
		      wrap (agentlookup_cb, u->root, name, dpt, resp));
  }
  return dpt;
}

usrinfo::usrinfo (sfs_aid a)
  : aid (a), root (afsusrdir::alloc (afs_root)), clonelock (false)
{
  root->link (afsaidfile::alloc (aid, strbuf () << aid << "\n"), ".aid");
  daemontab.traverse (wrap (flushuser, aid));
}

usrinfo::~usrinfo ()
{
  daemontab.traverse (wrap (flushuser, aid));
}

void
usrinfo::dispatch (svccb *sbp)
{
  if (!sbp) {
    x = NULL;
    sh = NULL;
    ch = NULL;
    usrtab.remove (aid);
    return;
  }

  switch (sbp->proc ()) {
  case AGENT_KILLSTART:
    root = afsusrdir::alloc (afs_root);
    root->link (afsaidfile::alloc (aid, strbuf () << aid << "\n"), ".aid");
    /* above implies root->bumpmtime (); */
    /* cascade */
  case AGENT_START:
    daemontab.traverse (wrap (flushuser, aid));
    sbp->replyref (0);
    break;
  case AGENT_KILL:
    sbp->replyref (0);
    usrtab.remove (aid);
    break;
  case AGENT_SYMLINK:
    {
      sfsagent_symlink_arg *ssa
	= sbp->template getarg<sfsagent_symlink_arg> ();
      root->mkulink (ssa->contents, ssa->name);
      sbp->reply (NULL);
      break;
    }
  case AGENT_FLUSHNAME:
    {
      str name = *sbp->template getarg<sfs_filename> ();
      root->clrulink (name);
      sbp->reply (NULL);
      break;
    }
  case AGENT_FLUSHNEG:
    {
      root->clrnegcache ();
      sbp->reply (NULL);
      break;
    }
  default:
    agent_default (sbp, aid);
    break;
  }
}

static void
usrinfo_accept (ptr<axprt_unix> ax, const authunix_parms *aup)
{
  if (!ax)
    fatal << "agent.sock" << ": " << strerror (errno) << "\n";
  assert (!ax->ateof ());
  vNew newagent (ax, aup);
}

void
usrinfo_init ()
{
  sfs_suidserv ("agent", wrap (usrinfo_accept));
}

static void
unlinkdir (str dirname, const sfs_aid &, ptr<usrinfo> u)
{
  /* Things that aren't directories are revocations -- leave them around */
  if (afsnode *e = u->root->lookup (dirname, NULL))
#if FIX_MNTPOINT
    if (opt_fix_mntpoint) {
      if (typeid (*e) != typeid (revocation))
	u->root->unlink (dirname);
    }
    else
#endif /* FIX_MNTPOINT */
      if (e->type == NF3DIR)
	u->root->unlink (dirname);
}

void
flushpath (str path)
{
  if (afsnode *e = afs_naroot->lookup (path, NULL))
#if FIX_MNTPOINT
    if (opt_fix_mntpoint) {
      if (typeid (*e) != typeid (revocation))
	afs_naroot->unlink (path);
    }
    else
#endif /* FIX_MNTPOINT */
      if (e->type == NF3DIR)
	afs_naroot->unlink (path);
  usrtab.traverse (wrap (unlinkdir, path));
}
